@@ -5,6 +5,7 @@ os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
 import tensorflow as tf
 import SimpleITK as sitk
 import json
+import numpy as np
 import sys
 
 AUTOTUNE = tf.data.experimental.AUTOTUNE
@@ -29,9 +30,9 @@ class DataLoader:
     def __init__(
         self,
         mode,
-        input_size=None,
         imgA_label=None,
         imgB_label=None,
+        input_size=None,
         data_dir="./Data",
         output_dir=None,
         is_B_categorical=False,
@@ -39,40 +40,50 @@ class DataLoader:
         norm_boundsA=None,
         norm_boundsB=None,
         extract_only=None,
+        use_3D=False,
     ):
         """[summary]
 
         Args:
+            mode ([type]): [description]
             imgA_label (str): Identifier for class A. It's the name of the
-                folder inside :py:attr:`imgs_subdir` that contains images
+                folder inside :py:attr:`data_dir` that contains images
                 labeled as class A.
             imgB_label (str): Identifier for class B. It's the name of the
-                folder inside :py:attr:`imgs_subdir` that contains images
+                folder inside :py:attr:`data_dir` that contains images
                 labeled as class B.
+            input_size (int): Dimension of a single image, defined as
+                input_size x input_size. Currently, it supports only squared
+                images.
             data_dir (str, optional): Path to directory that contains the
-                Dataset. This folder **must** contain a subfolder named like
-                :py:attr:`imgs_subdir`. Defaults to './Data'.
-            imgs_subdir (str, optional): Name (**not** the entire path) of the
-                folder that actually contains :py:attr:`imgA_label` and
-                :py:attr:`imgB_label` subfolders. It's a subfolder of Defaults
-                to 'Images'.
-            cache_mode (str, optional): One between "prod" or "test". "test" is
-                used as a debug modality only in case you want to create Cache
-                directory at the current folder location. "prod" creates a
-                subfolder named "Cache" inside :py:attr:`data_dir`. Defaults to
-                "prod".
+                Dataset. This folder **must** contain two subfolders named like
+                :py:attr:`imgA_label` and :py:attr:`imgB_label`. Defaults to
+                './Data'.
+            output_dir ([type], optional): [description]. Defaults to None.
+            is_B_categorical (bool, optional): [description]. Defaults to False.
+            num_classes ([type], optional): [description]. Defaults to None.
+            norm_boundsA ([type], optional): [description]. Defaults to None.
+            norm_boundsB ([type], optional): [description]. Defaults to None.
             extract_only (int, optional): Indicate wheter to partially cache a
                 certain amount of elements in the dataset. Please remember that
-                if "Cache" folder is already populated, you need to clean this
-                folder content to recreate a partial cache file. When it is
-                set to None, the entire Dataset is cached. Defaults to None.
+                if :py:attr:`output_dir` folder is already populated, you need
+                to clean this folder content to recreate a partial cache file.
+                When it is set to None, the entire Dataset is cached. Defaults
+                to None.
+            use_3D: Indicate whether to use three-dimensional data in the cache
+                (if True) or to extract two-dimensional slices from the 3D
+                volumes (if False). Defaults to False.
 
         Raises:
-            FileNotFoundError: :py:attr:`data_dir` doesn't exists.
-            FileNotFoundError: :py:attr:`imgs_subdir` doesn't exists.
-            FileNotFoundError: :py:attr:`imgA_label` doesn't exists.
-            FileNotFoundError: :py:attr:`imgB_label` doesn't exists.
-            ValueError: :py:attr:`cache_mode` is not "prod" nor "test".
+            ValueError: [description]
+            FileNotFoundError: [description]
+            ValueError: [description]
+            FileNotFoundError: [description]
+            FileNotFoundError: [description]
+            ValueError: [description]
+            ValueError: [description]
+            ValueError: [description]
+            FileNotFoundError: [description]
         """
 
         if mode not in __dataloader_modality__:
@@ -98,6 +109,10 @@ class DataLoader:
 
             self.input_size = input_size
 
+            if (not isinstance(use_3D, bool)):
+                raise ValueError("use_3D is not a Boolean value")
+            self.use_3D = use_3D
+
             if imgA_label is None or imgB_label is None:
                 raise ValueError("imgA_label or imgB_label is None.")
 
@@ -127,6 +142,11 @@ class DataLoader:
             else:
                 self.is_A_RGB = False
                 self.is_B_RGB = False
+
+            if ((not self.is_3D) and (self.use_3D)):
+                raise ValueError(
+                    "Image files are not 3D but use_3D was set to True")
+
             self.imgA_type = self.check_type(self.imgA_paths[0])
             self.imgB_type = self.check_type(self.imgB_paths[0])
             self.is_B_categorical = is_B_categorical
@@ -158,6 +178,7 @@ class DataLoader:
                                 "num_classes": self.num_classes,
                                 "norm_boundsA": self.norm_boundsA,
                                 "norm_boundsB": self.norm_boundsB,
+                                "use_3D": self.use_3D
                                 }
 
             output_dir_content = os.listdir(self.output_dir)
@@ -204,6 +225,7 @@ class DataLoader:
                 self.num_classes = dataset_property["num_classes"]
                 self.norm_boundsA = dataset_property["norm_boundsA"]
                 self.norm_boundsB = dataset_property["norm_boundsB"]
+                self.use_3D = dataset_property["use_3D"]
 
     def get_dataset(self,
                     batch_size=32,
@@ -253,10 +275,10 @@ class DataLoader:
                  norm_bounds=None):
         """Open image files for one class and store it inside cache.
 
-        This function performs all the(usually) slow reading operations that
+        This function performs all the (usually) slow reading operations that
         is necessary to execute at least the first time. After the first
         execution information are saved inside some cache file inside Cache
-        folder(typically created in your Dataset folder, at the same level of
+        folder (typically created in your Dataset folder, at the same level of
         Images folder). This function detects if cache files are already
         present, and in that case it skips the definition of these files.
         Please take into account that cache files will be as big as your
@@ -275,32 +297,40 @@ class DataLoader:
         index_file = f"{cache_file}.index"
 
         ds = tf.data.Dataset.from_tensor_slices(img_paths)
+
         ds = ds.map(lambda path: tf.py_function(self.open_img,
                                                 [path],
                                                 [__dict_dtype__[img_type]],
                                                 ),
                     num_parallel_calls=AUTOTUNE)
 
-        if norm_bounds is not None:
-            ds = ds.map(lambda img: self.norm_with_bounds(img,
-                                                          norm_bounds),
-                        num_parallel_calls=AUTOTUNE
-                        )
-
-        if self.is_3D:
+        if self.is_3D and (not self.use_3D):
             ds = ds.unbatch()
 
         if is_RGB:
-            ds = ds.map(lambda img: tf.image.rgb_to_grayscale(img))
+            ds = ds.map(lambda img: tf.image.rgb_to_grayscale(img),
+                        num_parallel_calls=AUTOTUNE)
 
-        ds = ds.map(lambda img: self.check_dims(img,
-                                                self.input_size),
+        # ds = ds.map(lambda img: self.check_dims(img,
+        #                                        self.input_size),
+        #            num_parallel_calls=AUTOTUNE)
+        ds = ds.map(lambda img: self.fix_image_dims(img,
+                                                    self.input_size),
                     num_parallel_calls=AUTOTUNE)
 
         if is_categorical:
             ds = ds.map(lambda img: tf.one_hot(tf.squeeze(tf.cast(img,
                                                                   img_type)),
-                                               depth=num_classes))
+                                               depth=int(num_classes)),
+                        num_parallel_calls=AUTOTUNE)
+
+        ds = ds.map(lambda img: tf.cast(img, img_type),
+                    num_parallel_calls=AUTOTUNE)
+
+        if norm_bounds is not None:
+            ds = ds.map(lambda img: self.norm_with_bounds(img,
+                                                          norm_bounds),
+                        num_parallel_calls=AUTOTUNE)
 
         ds = ds.cache(cache_file)
 
@@ -356,6 +386,9 @@ class DataLoader:
         path = path.numpy().decode("utf-8")
         image = sitk.GetArrayFromImage(sitk.ReadImage(path))
 
+        if (self.use_3D):
+            image = np.transpose(image, axes=(2, 1, 0))
+
         tensor = tf.convert_to_tensor(image)
 
         return tensor
@@ -396,19 +429,35 @@ class DataLoader:
 
         img_type = image.dtype.name
 
-        # return __dict_dtype__[img_type]
         return img_type
 
-    @ staticmethod
-    def check_dims(img, size):
-        img = tf.expand_dims(tf.squeeze(img), axis=-1)
-        img = tf.image.resize_with_pad(img, size, size)
+    def fix_image_dims(self, img, size):
+        """Fix tensor dimensions so that they are of the
+        proper size to carry out Tensorflow operations.
 
+        This function performs three steps:
+
+        #. `Squeeze <https://www.tensorflow.org/api_docs/python/tf/squeeze>`_ to remove axis with dimension of 1
+        #. `Expand <https://www.tensorflow.org/api_docs/python/tf/expand_dims>`_ the dimensions of the tensor by adding one axis
+        #. `Resize and pad <https://www.tensorflow.org/api_docs/python/tf/image/resize_with_pad>`_ the tensor to a target width and height
+
+        If `use_3D` was enabled, volume is not resized and padded.
+
+        Args:
+            img: image or volume to be processed
+            size: desired size of image or volume in the two/three axis.
+
+
+        """
+        img = tf.expand_dims(tf.squeeze(img), axis=-1)
+        if (not self.use_3D):
+            img = tf.image.resize_with_pad(img, size, size)
         return img
 
     # -------------------------------------------------------------------------
     #  Transformations
     # -------------------------------------------------------------------------
+
     @ staticmethod
     def norm_with_bounds(image, bounds):
         """Image normalisation. Normalises image in the range defined by lb and
@@ -416,8 +465,8 @@ class DataLoader:
         lb = tf.cast(bounds[0], dtype=image.dtype)
         ub = tf.cast(bounds[1], dtype=image.dtype)
 
-        tf.where(image < lb, lb, image)
-        tf.where(image > ub, ub, image)
+        image = tf.where(image < lb, lb, image)
+        image = tf.where(image > ub, ub, image)
 
         image = image - lb
         image /= (ub - lb)
@@ -460,6 +509,7 @@ def generate_dataset(data_dir,
                      norm_boundsB=None,
                      is_B_categorical=False,
                      num_classes=None,
+                     use_3D=False,
                      ):
 
     data_loader = DataLoader(mode="gen",
@@ -473,6 +523,7 @@ def generate_dataset(data_dir,
                              norm_boundsA=norm_boundsA,
                              norm_boundsB=norm_boundsB,
                              extract_only=extract_only,
+                             use_3D=use_3D
                              )
 
     data_loader.get_dataset()
@@ -486,7 +537,7 @@ def get_dataset(data_dir,
                 train_augmentation=True,
                 random_crop_size=None,
                 random_rotate=True,
-                random_flip=True,
+                random_flip=True
                 ):
 
     if len(percentages) != 3:
